var SOCKET_HOST = 'ws://localhost:9030';
var SOCKET_RECONNECT_PERIOD = 2000;

var STATUS_SPACING = 15;
var STATUS_PADDING_TOP = 15;
var STATUS_PADDING_LEFT = 10;
var STATUS_WIDTH = 150;
var STATUS_HEIGHT = 200;
var STATUS_FPS_PERIOD = 1000;

var MOVE_SPACING = 0.1;

var ANIMATE_STEP = 500;
var TOTAL_STEP = 700;

var gridSize = 20;
var gridColor = 'rgba(0,0,0,0.1)';